package org.greenstone.gsdl3.gs3build.doctypes;

import java.util.List;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.net.URL;

import org.greenstone.gsdl3.gs3build.metadata.NamespaceFactory;
import org.greenstone.gsdl3.gs3build.metadata.GSDL3Namespace;
import org.greenstone.gsdl3.gs3build.metadata.METSDescriptiveSet;
import org.greenstone.gsdl3.gs3build.metadata.METSFile;
import org.greenstone.gsdl3.gs3build.metadata.METSFileSet;
import org.greenstone.gsdl3.gs3build.metadata.METSHeader;
import org.greenstone.gsdl3.gs3build.metadata.METSStructure;
import org.greenstone.gsdl3.gs3build.metadata.METSStructureSet;
import org.greenstone.gsdl3.gs3build.metadata.METSDivision;
import org.greenstone.gsdl3.gs3build.metadata.METSNamespace;

import org.greenstone.gsdl3.gs3build.util.MultiMap;

/**
 *  Provide a base-line functionality for the <code>DocumentInterface</code>
 *  class.
 */

public abstract class AbstractDocument implements DocumentInterface
{
  METSFileSet      fileSet;
  METSDescriptiveSet metadata;
  METSStructureSet structureSet;
  METSHeader       header;
  DocumentID       id;
  
  public AbstractDocument(URL url)
  { this.fileSet      = new METSFileSet();
    METSFile metsFile = this.fileSet.addFile(url);
    this.metadata     = new METSDescriptiveSet();
    this.header       = new METSHeader();
    this.structureSet = new METSStructureSet();

    METSStructure structure = new METSStructure("All", "All", "Whole Document");
    METSDivision documentBody = new METSDivision("All", "All", "All", "Whole Document", "Document");
    structure.addDivision(documentBody);
    this.structureSet.addStructure(structure);
    documentBody.addFileReference(metsFile.getID());
    documentBody.addMetadataReference("DM1");
  }

  /**
   *  Set the identified for the document.  Every document should have
   *  a document number set on its accession, either through metadata
   *  placed upon it internally or externally, or by assignment through
   *  a <code>DocumentIDFactory</code>.  Each identifier should be
   *  unique.
   *
   *  @param <code>DocumentID</code> the document identifier - in XML
   *         terms, the gsdl3:id element.
   */
  public void setID(DocumentID id)
  { this.id = id;
  }

  /**
   *  Get the document identifier - this should be unique to the document,
   *  but care must be taken in the configuration of the collection to
   *  ensure that this is the case.
   *
   *  @return <code>DocumentID</code> the identifer
   */
  public DocumentID getID()
  { return this.id;
  }

  /**
   *  Indicate whether this document is indexed.
   *
   *  @see: DocumentInterface.isIndexed
   */
  public boolean isIndexed()
  {
    return true;
  }

  /**
   *  Obtain the METS header of this document
   *
   *  @return <code>METSHeader</code> the header
   */
  public METSHeader getHeader()
  { return this.header;
  }	

  /**
   *  Set the METS header for this document.
   *
   *  @param <code>METSHeader</code> the header
   */
  public void setHeader(METSHeader header)
  { this.header = header;
  }

  /**
   *  A simple implementation of the isDocumentType function that does <b>not</b> consider
   *  inheritance - it <code>must</code> be extended as required.
   */
  public boolean isDocumentType(String type)
  { return type.equals(this.getDocumentType());
  }

  public abstract String getDocumentType();
  
  public abstract String getDocumentText();

  public String getMETSType()
  { return "document";
  }

  /**
   *   @see DocumentInterface:addDocumentMetadata
   */
  public void addDocumentMetadata(String name, String value)
  { int colonAt = name.indexOf(":");
    String namespace;
		
    if (colonAt > 0) {
      namespace = name.substring(0, colonAt);
      name = name.substring(colonAt+1);
    }
    else {
      namespace = GSDL3Namespace.GSDL3_NAMESPACE_ID;
    }
    this.addDocumentMetadata(namespace, name, value);
  }

  /**
   *   @see DocumentInterface:addDocumentMetadata
   */
  public void addDocumentMetadata(String namespace, String label, String value)
  { this.metadata.addMetadata("default", namespace, label, value);
  }

  /**
   *  @see DocumentInterace:setDocumentMetadata
   */
  public void setDocumentMetadata(String namespace, String label, String value)
  { this.metadata.setMetadata("default", namespace, label, value);
  }

  /**
   *  Get the metadata structure of the document
   *
   *  @return <code>METSDescriptive</code> the metadata holder for the document.
   */
  public METSDescriptiveSet getDocumentMetadata()
  { return this.metadata;
  }

  /**
   *  Get the metadata structure of the document
   *
   *  @return <code>METSStructureSet</code> the metadata holder for the document.
   */
  public METSStructureSet getDocumentStructure()
  { return this.structureSet;
  }

  /**
   *  Get the values associated with a particular metadata value.
   *
   *  @param <code>String</code> the namespace to find the values in.
   *  @param <code>String</code> the label to match to find the values.
   *
   *  @return <code>List</code> the values.
   */
  public List getDocumentMetadataItem(String namespace, String label)
  { return this.metadata.getMetadata("default", namespace, label);
  }

  /**
   *  Get the values associated with a particular metadata value.
   *
   *  @param <code>String</code> the namespace and label separated by a
   *         colon.
   *
   *  @return <code>List</code> the values.
   */
  public List getDocumentMetadataItem(String namespaceLabel)
  { String namespace, label;
  
    int colonAt = namespaceLabel.indexOf(':');
    if (colonAt < 0)
    { namespace = GSDL3Namespace.GSDL3_NAMESPACE_ID;
      label     = namespaceLabel;
    }
    else
    { namespace = namespaceLabel.substring(0, colonAt);
      label     = namespaceLabel.substring(colonAt+1);
    }
    return this.metadata.getMetadata("default", namespace, label);
  }

  /**
   *  @see DocumentInterface:getDocumentFiles
   */
  public METSFileSet getDocumentFiles()
  { return this.fileSet;
  }
  
  /**
   *  @see DocumentInterface:isMETSCompatible
   */
  public boolean isMETSCompatible()
  { return true;
  }

  /**
   *  Use a default document writer - this may be overridden for subclasses...
   *
   *  @see DocumentInterface:writeMETSObject
   */
  public DocumentWriter getMETSWriter()
  { return new DocumentWriter();
  }

  /**
   *  Use a default SQL document writer - this may be overridden for subclasses...
   *
   */
  public DocumentSQLWriter getSQLWriter()
  { return new DocumentSQLWriter();
  }
    
}
