package org.greenstone.gsdl3.gs3build.metadata;

import java.io.PrintWriter;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;

import java.sql.SQLException;

import org.greenstone.gsdl3.gs3build.util.XMLTools;
import org.greenstone.gsdl3.gs3build.util.GS3SQLConnection;
import org.greenstone.gsdl3.gs3build.database.*;


public class METSDivision
{
  String       ID;
  String       order;       // see the METS documentation for the meaning of the different
  String       orderLabel;  // label types.
  String       userLabel;
  String       type;        // e.g. section, chapter, physical or logical, etc.
  List         metadataRefs;
  List         adminRefs;
  List         fileRefs;
  Map          children;

  public METSDivision(String ID, String order, String orderLabel, String userLabel, String type)
  { this.ID = ID;
    this.order = order;
    this.orderLabel = orderLabel;
    this.userLabel = userLabel;
    this.type = type;
    this.fileRefs = new ArrayList();
    this.metadataRefs = new ArrayList();
    this.children = new HashMap();
  }

  public String getID()
  { return this.ID;
  }

	/**
	 *  Add a new file reference to the structure
	 *
	 *  @param <code>METSFileRef</code> the file to be referenced.
	 */
	public void addFileReference(METSFileID fileRef)
	{	this.fileRefs.add(fileRef);
	}

    /**
     *  Add a new metadata reference to the structure
     */
    public void addMetadataReference(String metadataRef)
    { // TODO: correct the type
      this.metadataRefs.add(metadataRef);
    }

	/**
	 *  Add a child division, or sub-division to the document.  For example,
	 *  sections within a chapter.
	 *
	 *  @param <code>METSDivision</code> the child division.
	 */
	public void addSubDivision(METSDivision child)
	{	this.children.put(child.getID(), child);
	}

	/**
	 *  Add a child division to a descendent of this division.
	 *
	 *  @param <code>String</code> the path to the division into which the
	 *         subdivision will be added.
	 *  @param <code>METSDivision</code> the division to be added
	 */
	public void addSubDivision(String divisionID, METSDivision division)
	{	String topDivision = METSDivision.getTopDivisionName(divisionID);

		METSDivision child = (METSDivision) this.children.get(topDivision);
		if (child == null)
		{	return;
		}

		String subDivisionID = METSDivision.getSubDivisionName(divisionID);
		if (subDivisionID == null)
		{ child.addSubDivision(division);
		}
		else
		{	child.addSubDivision(subDivisionID, division);
		}
	}

	public METSDivision getDivision(String divisionID)
	{	String topDivision = METSDivision.getTopDivisionName(divisionID);

		METSDivision child = (METSDivision) this.children.get(topDivision);
		if (child == null)
		{	return null;
		}

		String subDivision = METSDivision.getSubDivisionName(divisionID);
		if (subDivision != null)
		{	child = child.getDivision(subDivision);
		}

		return child;
	}

	public static String getSubDivisionName(String divisionName)
	{	int at = divisionName.indexOf('.');

		if (at >= 0)
		{	return divisionName.substring(at+1);
		}
		else
		{	return null;
		}
	}

	public static String getTopDivisionName(String divisionName)
	{	int at = divisionName.indexOf('.');

		if (at >= 0)
		{	return divisionName.substring(0, at);
		}
		else
		{	return divisionName;
		}
	}

    /**
     *  Write the METS file in an XML to a text-output sink
     *
     *  @param <code>PrintWriter</code> the destination of the output
     */
    public void write(PrintWriter writer)
    { String tag = XMLTools.getOpenTag("mets", "div");
      tag = XMLTools.addAttribute(tag, "ID", this.ID.toString());
      tag = XMLTools.addAttribute(tag, "TYPE", this.type);
      tag = XMLTools.addAttribute(tag, "ORDER", this.order);
      if (this.orderLabel != null)
      { tag = XMLTools.addAttribute(tag, "ORDERLABEL", this.orderLabel);
      }
      if (this.userLabel != null)
      { tag = XMLTools.addAttribute(tag, "LABEL", this.userLabel);
      }
      writer.println(tag);

      // write the list of file pointers for this structural element
      if (this.fileRefs.size() > 0)
      {	tag = XMLTools.getOpenTag("mets", "fptr");
      
        String fileList = this.writeList(this.fileRefs);

	tag = XMLTools.addAttribute(tag, "FILEID", fileList);

	tag = XMLTools.makeSingleton(tag);
	writer.println(tag);
      }

      // ..and then the metadata - this is the same code at present,
      // but is replicated here in case of future amendment.
      if (this.metadataRefs.size() > 0)
      { tag = XMLTools.getOpenTag("mets", "mptr");

        String metaList = this.writeList(this.metadataRefs);
        tag = XMLTools.addAttribute(tag, "METAID", metaList);
	tag = XMLTools.makeSingleton(tag);
	writer.println(tag);
      }
      
      // close the div element
      writer.println(XMLTools.getCloseTag("mets", "div"));
    }

    private String writeList(List list)
    { StringBuffer listString = new StringBuffer();
      
      Iterator iterator = list.iterator();
      while (iterator.hasNext())
      { String item = iterator.next().toString();
	if (listString.length() > 0)
	{ listString.append(" ");
	}
	listString.append(item);
      }
      return listString.toString();
    }

    public void writeSQL(int parentRef, boolean parentIsStructure, GS3SQLConnection connection)
    { Iterator groups = this.children.values().iterator();

      GS3SQLInsert insert = new GS3SQLInsert("divisions");
      insert.addValue("SectionID", this.ID.toString());
      insert.addValue("Type", this.type.toString());
      insert.addValue("LabelOrder", this.order);
      insert.addValue("ShortLabel", this.orderLabel);
      insert.addValue("UserLabel", this.userLabel);
      insert.addValue("ParentRef", Integer.toString(parentRef));
      insert.addValue("ParentType", parentIsStructure == true ? "Structure" : "Division");
      
      System.out.println(insert.toString());
      connection.execute(insert.toString());

      // TODO: write metadata references
      // TODO: write file references
      
      // get the new structure reference
      GS3SQLSelect select = new GS3SQLSelect("divisions");
      select.addField("divisionRef");
      GS3SQLWhereItem whereItem = new GS3SQLWhereItem("ParentRef", "=", Integer.toString(parentRef), GS3SQLField.INTEGER_TYPE);
      GS3SQLWhereItem parentItem = new GS3SQLWhereItem("ParentType", "=", parentIsStructure ? "Structure" : "Division");
      GS3SQLWhereItem item = new GS3SQLWhereItem("SectionID", "=", this.ID.toString());
      GS3SQLWhere where = new GS3SQLWhere(whereItem);
      where.add(parentItem);
      where.add(item);
      select.setWhere(where);
      connection.execute(select.toString());

      // get the sql reference for the division
      int sqlRef;

      try {
        // read in the structure reference
	connection.getResultSet().first();
	sqlRef = connection.getResultSet().getInt("divisionRef");
      }
      catch (SQLException sqlex) {
	  System.out.println("Unable to retrieve reference for Division " + sqlex);
	  return;
      }
        
      // write the file references
      if (this.fileRefs.size() > 0)
      { Iterator iterator = this.fileRefs.iterator();
	
        while (iterator.hasNext()) 
	{ GS3SQLInsert fileinsert = new GS3SQLInsert("divisionfilerefs");
	  fileinsert.addValue("divisionRef", Integer.toString(sqlRef), GS3SQLField.INTEGER_TYPE);
	  fileinsert.addValue("Type", "Group");
	  fileinsert.addValue("fileRef", iterator.next().toString());
	  connection.execute(fileinsert.toString());
	}
      }
        
      // write the metadata references
      if (this.metadataRefs.size() > 0)
      { Iterator iterator = this.metadataRefs.iterator();
	
        while (iterator.hasNext()) 
	{ GS3SQLInsert metainsert = new GS3SQLInsert("divisionmetarefs");
	  metainsert.addValue("divisionRef", Integer.toString(sqlRef), GS3SQLField.INTEGER_TYPE);
	  metainsert.addValue("Type", "Group");
	  metainsert.addValue("metadataRef", iterator.next().toString());
	  connection.execute(metainsert.toString());
	}
      }

      // write out any children in turn
      while (groups.hasNext())
      { METSDivision group = (METSDivision) groups.next();

        group.writeSQL(sqlRef, false, connection);
      }
    }

}
