package org.greenstone.gsdl3.gs3build.metadata;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import java.io.PrintWriter;

import org.greenstone.gsdl3.gs3build.util.XMLTools;
import org.greenstone.gsdl3.gs3build.util.GS3SQLConnection;
import org.greenstone.gsdl3.gs3build.database.*;

import org.greenstone.gsdl3.gs3build.doctypes.DocumentInterface;


public class METSFileGroup
{ List   children;
  List   childGroups;
  String id;

  public METSFileGroup(String id)
  { this.children = new ArrayList();
    this.childGroups = new ArrayList();
    this.id = id;
  }

	/**
	 *  Add a filegroup to this group of files.
	 *
	 *	@param <code>METSFileGroup</code> the filegroup.
	 */
	public void addGroup(METSFileGroup group)
	{	this.childGroups.add(group);
	}

	/**
	 *  Add a file to this group of files.
	 *
	 *	@param <code>METSFile</code> the file in a METS File object.
	 */
	public void addFile(METSFile file)
	{	this.children.add(file);
		file.setGroup(this);
		if (!file.getID().isDefined())
		{ file.setID(new METSFileID(this.id + Integer.toString(this.children.size())));
		}
	}

	/**
	 *  Get the nth child of the group
	 *
	 *  @param <code>int</code> the index into the group
	 *  @return <code>METSFile</code> the corresponding child.  This may be <code>null</code>
	 *          particularly if the index given falls outside of the valid range of child
	 *          indexes.
	 */
	public METSFile getFile(int index)
	{ if (index < 0 || index >= this.children.size())
		{	return null;
		}

		return (METSFile) this.children.get(index);
	}

	/**
	 *  Get all the files in the group as a <code>List</code>
	 */
	public List getFiles()
	{	return this.children;
	}

	/**
	 *  Get all the subgroups as a <code>List</code>
	 */
	public List getSubgroups()
	{	return this.childGroups;
	}

	/**
	 *  Get the number of files in the group
	 *
	 *  @return <code>int</code> the count
	 */
	public int noOfFiles()
	{	return this.children.size();
	}

	/**
	 *  Get the number of subgroups.
	 */
	public int noOfSubgroups()
	{	return this.childGroups.size();
	}

	/**
	 *  @return <code>String</code> the name of this group
	 */
	public String getName()
	{	return this.id;
	}

  /**
   *  Write the object in an XML METS format.
   *
   *  @param <code>PrintWriter</code> the printwriter being used to write the object
   */
  public void write(PrintWriter writer)
  { String tag = XMLTools.getOpenTag("mets","fileGrp");
    tag = XMLTools.addAttribute(tag, "ID", this.id);
    writer.println(tag);
		
    Iterator childIter = children.iterator();
    while (childIter.hasNext())
    { METSFile file = (METSFile) childIter.next();
      file.write(writer);
    }
    writer.println(XMLTools.getCloseTag("mets","fileGrp"));
  }

  public void writeSQL(DocumentInterface document, GS3SQLConnection connection)
  { // check if this node is in the
    GS3SQLSelect select = new GS3SQLSelect("filegroups");
    select.addField("*");
    GS3SQLWhereItem whereItem = new GS3SQLWhereItem("FileGroupID", "=", this.id);
    GS3SQLWhereItem whereDoc  = new GS3SQLWhereItem("DocID", "=", document.getID().toString());
    GS3SQLWhere where = new GS3SQLWhere(whereItem);
    where.add(whereDoc);
    select.setWhere(where);

    System.out.println(select.toString());
    connection.execute(select.toString());

    GS3SQLInsert insert = new GS3SQLInsert("filegroups");
    insert.addValue("DocID", document.getID().toString());
    insert.addValue("FileGroupID", this.id);
    insert.addValue("ParentGroup", "");
    System.out.println("filegroup done");
    connection.execute(insert.toString());
		
    System.out.println("filegroup done");

    Iterator childIter = children.iterator();
    while (childIter.hasNext())
    { METSFile file = (METSFile) childIter.next();
      file.writeSQL(connection);
    }
  }
}
